/***************************************************************************
                       inv_folding.cpp  -  main
                             -------------------
    begin                : 30-08-2004
    copyright            : (C) 2004 by Anke Busch
    email                : anke.busch@informatik.uni-freiburg.de
 ***************************************************************************/

#include <iostream>
#include <stdlib.h>

#include "basics.h"
#include "struct.h"
#include "inverse.h"
#include "search.h"

using namespace std;


/**********************************************************************************
*    Constants                                                                   *
*          defined in "basics.h"
**********************************************************************************/

const int MAX_LOOPS = 2000; //max. number of loops

const double T                    = 310.15  ; // temperature in Kelvin(37C);K=C+273.15
const double RT                   = 0.616   ; // Boltzmann gas constant in kcal/mol
// RT=kT=1.380622*10^(-23)Joule/K *310.15K (37C=273.15K+37K=310.15K)

const double MIN_DOUBLE = -1000000.0;
const double MAX_DOUBLE =  1000000.0;
const double DOUBLE_DIFF = 0.00001;

const int MIN_INT = -1000000;
const int MAX_INT = 1000000;

/**********************************************************************************
*                               global variabels                                  *
**********************************************************************************/

char* brackets;  //Structure given in bracket notation
int struct_len;  //size of the structure
char *ElementStruct;  // another representation of the structure (incl. elements like S, M, B,...)
int** BP_Order;  //order of the base pairs (BPs) in which they are examined
int numBP;       //number of base pairs
int*  BP_Pos_Nr; //vector in which for each pos. in brackets the pos. in BP_Order is stored

double** D;    //Dynamic recursion matrix
int**** Trace; //Dynamic Traceback Matrix (has 2 dimensions more than D, since in each D-corresponding field the coordinates of all
               // predecessor fields have to be stored (two coord. per predecessor) and since there could be more than one predecessors
               // 2 more dimensions are needed)

char* best_char_seq;   // designed sequence
bool free_bases_set2A; // is true, if free bases have to be set to 'A'
bool free_bases_set2C;
bool free_bases_set2G;
bool free_bases_set2U;

int search_strategy;       // gives the search strategy ( 1 = adaptive walk, 2 = full local search, 3 = stochastic local search)
int neighbour_choice;      // gives the kind of ranking of the neighbors (1 = random, 2 = energy dependent)
int only_mutation_is_step; // gives the information how to count the step done during the stochastic local search
int step_multiplier;       // maximal number of steps during SLS = step_multiplier * length
double p_accept;           //probability to accept worse neighbors during SLS



/**********************************************************************************
*                        help for calling the program                             *
**********************************************************************************/

void usage(char* name)
{
   cout << "\ncall: " << name << " \"structure\" \n";
   cout << "                              [-f[ACGUMRWSYKVHDBN] assignment where free bases are set to]\n";
   cout << "                              [-F[mp]] [-R [repeats]] [-S search strategy] [-m]\n";
   cout << "                              [-s length multiplier for max. number of steps during SLS] \n";
   cout << "                              [-N neighbor choice] [-p prob. to accept worse neighbors]\n\n";
   exit(1);
}


void usage_help(char* name)
{
   cout << "\ncall: " << name << " \"structure\" \n";
   cout << "                              [-f[ACGUMRWSYKVHDBN] assignment where free bases are set to]\n";
   cout << "                              [-F[mp]] [-R [repeats]] [-S search strategy] [-m]\n";
   cout << "                              [-s length multiplier for max. number of steps during SLS] \n";
   cout << "                              [-N neighbor choice] [-p prob. to accept worse neighbors]\n\n";
   cout << endl;
   cout << "\nOptions of the initializing part:\n";
   cout << "------------------------------------\n";
   cout << " -f[ACGUMRWSYKVHDBN]\t Give the assignment (in IUPAC code) where free bases in loop\n";
   cout << " \t\t\t regions (that give no energy fraction) are set to.\n"; 
   cout << "\nOptions of the local search part:\n";
   cout << "-----------------------------------\n";
   cout << " -F[mp]\t\t Use minimum energy (-Fm) (default), partition function folding (-Fp)\n";
   cout << " \t\t or both (-Fmp).\n";
   cout << endl;
   cout << " -R[repeats]\t Number of repeating the local search step.\n";
   cout << endl;
   cout << " -S strategy\t Search strategy used during the local search step: \n";
   cout << "                            1 - adaptive walk\n";
   cout << "                            2 - full local search\n";
   cout << "                            3 - stochastic local search (default)\n";
   cout << endl;
   cout << " -m\t\t Kind of counting the step during the local search. Here, only accepted\n";
   cout << " \t\t mutations are counted, while usually all tested neighbors are counted.\n";
   cout << endl;
   cout << " -s multiplier\t The maximal number of step during the stochastic local search\n";
   cout << " \t\t is given as product of the structure length and a multiplier. The\n";
   cout << " \t\t multiplier is set to 10 by default.\n";
   cout << endl;
   cout << " -N neighbor choice\t Kind of order the candidate neighbors during the local search.\n";
   cout << "                            1 - randomly\n";
   cout << "                            2 - energy dependent (default)\n";
   cout << endl;
   cout << " -p probability\t Probability to accept worse neighbors during the stochastic local\n";
   cout << " \t\t search. It is set to 0.1 by default.\n";
   cout << endl;
   
   exit(1);
}

/**********************************************************************************
*         Print the input and the results of the initializing step                *
**********************************************************************************/

void print_in_and_output(double best_energy)
{
   printf("\n=========================\n");
   printf("Initializing Step: \n");
   printf("=========================\n");
   printf("Wanted Structure  : %s\n",brackets);
   printf("Designed Sequence : %s\n",best_char_seq);
   //printf("Energy: %1.2f\n", best_energy);
   printf("\n=========================\n");
   printf("Local Search Results: \n");
   printf("=========================\n");
}


/**********************************************************************************
*                                 Main-Function                                   *
**********************************************************************************/

int main(int argc, char *argv[])
{
   char *rstart, *str2;
   int hd, mfe = 1, pf = 0, repeat = 0, found;
   double energy = 0.0, kT;

   search_strategy = 3;
   neighbour_choice = 2;
   only_mutation_is_step = 0;
   step_multiplier = 10;
   p_accept = 0.1;

   do_backtrack = 0;

   free_bases_set2A = 0;
   free_bases_set2C = 0;
   free_bases_set2G = 0;
   free_bases_set2U = 0;

   if (argc < 2)
   {
      usage(argv[0]);
      exit(-1);
   }

   if ((argv[1][0] == '-') && (argv[1][1] == 'h'))
   {
      usage_help(argv[0]);
      exit(1);
   }
   else
      brackets = argv[1];

   for (int i = 2; i<argc; i++)
   {
      if (argv[i][0] == '-')
         switch (argv[i][1])
         {
            case 'f': //i++;
                      //if ( i<argc )
                      if ((int)strlen(argv[i]) > 2)
                      {
                         for(int j=2; j<(int)strlen(argv[i]); j++)
                            switch( argv[i][j] )
                            {
                               case 'A' :  free_bases_set2A = 1;
                                           //cout << endl << "free_bases_set2A: " << free_bases_set2A;
                                           break;
                               case 'C' :  free_bases_set2C = 1;
                                           //cout << endl << "free_bases_set2C: " << free_bases_set2C;
                                           break;
                               case 'G' :  free_bases_set2G = 1;
                                           //cout << endl << "free_bases_set2G: " << free_bases_set2G;
                                           break;
                               case 'U' :  free_bases_set2U = 1;
                                           //cout << endl << "free_bases_set2U: " << free_bases_set2U;
                                           break;
                               case 'M' :  free_bases_set2A = 1;
                                           free_bases_set2C = 1;
                                           break;
                               case 'R' :  free_bases_set2A = 1;
                                           free_bases_set2G = 1;
                                           break;
                               case 'W' :  free_bases_set2A = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'S' :  free_bases_set2C = 1;
                                           free_bases_set2G = 1;
                                           break;
                               case 'Y' :  free_bases_set2C = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'K' :  free_bases_set2G = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'V' :  free_bases_set2A = 1;
                                           free_bases_set2C = 1;
                                           free_bases_set2G = 1;
                                           break;
                               case 'H' :  free_bases_set2A = 1;
                                           free_bases_set2C = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'D' :  free_bases_set2A = 1;
                                           free_bases_set2G = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'B' :  free_bases_set2C = 1;
                                           free_bases_set2G = 1;
                                           free_bases_set2U = 1;
                                           break;
                               case 'N' :  free_bases_set2A = 1;
                                           free_bases_set2C = 1;
                                           free_bases_set2G = 1;
                                           free_bases_set2U = 1;
                                           break;
                               default : usage(argv[0]);
                            }
                         //cout << endl << "free_bases_set2A: " << free_bases_set2A;
                         //cout << endl << "free_bases_set2C: " << free_bases_set2C;
                         //cout << endl << "free_bases_set2G: " << free_bases_set2G;
                         //cout << endl << "free_bases_set2U: " << free_bases_set2U;
                      }
                      else
                         usage(argv[0]);
                      break;
            case 'F': mfe = 0; pf = 0;
                      for(int j=2;j<(int)strlen(argv[i]);j++)
                      {
                         switch( argv[i][j] )
                         {
                            case 'm' :  mfe = 1;
                                        break;
                            case 'p' :  pf = 1; /* old version had dangles=0 here */
                                        break;
                            default : usage(argv[0]);
                         }
                      }
                      break;
            case 'R': repeat = 100; //REPEAT_DEFAULT;
                      if(++i<argc)
                         if (sscanf(argv[i], "%d", &repeat)==0)
                            usage(argv[0]);
                      break;
            case 'S': if (argv[i][2]!='\0')
                         usage(argv[0]);
                      if (sscanf(argv[++i], "%d", &search_strategy)==0)
                         usage(argv[0]);
                      break;
            case 'N': if (argv[i][2]!='\0')
                         usage(argv[0]);
                      if (sscanf(argv[++i], "%d", &neighbour_choice)==0)
                         usage(argv[0]);
                      break;
            case 's': if (argv[i][2]!='\0')
                         usage(argv[0]);
                      if (sscanf(argv[++i], "%d", &step_multiplier)==0)
                         usage(argv[0]);
                      break;
            case 'p': if (argv[i][2]!='\0')
                         usage(argv[0]);
                      //p_accept = atof(argv[++i]);
                      if (sscanf(argv[++i], "%lf", &p_accept)==0)
                         usage(argv[0]);
                      break;
            case 'm': only_mutation_is_step = 1;
                      break;
            default : usage(argv[0]);
         }
      else
         usage(argv[0]);
   }


   if ((search_strategy > 3) || (search_strategy < 1))
   {
      printf("The search strategy is not valid.\n");
      exit(1);
   }

   if ((neighbour_choice > 2) || (neighbour_choice < 1))
   {
      printf("The choice of the neighbours is not valid.\n");
      exit(1);
   }

   //test, whether the structure is valid:
   //***************************************
   int correct = check_brackets(brackets);
   numBP = NumOfBP(brackets);
   struct_len = (int)strlen(brackets);

   if (correct)
   {
      getOrder(brackets);
      Pos2BP_Pos();
   }
   else
   {
      cerr << "\nNo valid structure!\n\n";
      exit(1);
   }

   //****************************************************
   //****************************************************

   //Initialization:
   //***************
   double x;
   x = Recursion();

   print_in_and_output(x);

   //****************************************************
   //****************************************************

   //Local Search:
   //*************

   init_rand();
   kT = (temperature+273.15)*1.98717/1000.0;
   give_up = (repeat<0);

   str2 = (char *) malloc(sizeof(char)*((unsigned)struct_len+1));

   if (repeat!=0) found = (repeat>0)? repeat : (-repeat);
   else found = 1;

   initialize_fold(struct_len);
   rstart = (char *) malloc(sizeof(char)*((unsigned)struct_len+1));

   while(found>0) 
   {
      char *string;
      string = (char *) malloc(sizeof(char)*((unsigned)struct_len+1));
      strcpy(string, best_char_seq);
      strcpy(rstart, string); /* remember start string */

      if (mfe)
      {
         energy = inverse_fold(string);
         if( (repeat>=0) || (energy<=0.0) )
         {
            found--;
            hd = hamming(rstart, string);
            printf("MFE:    %s  %3d", string, hd);
            if (energy>0) /* no solution found */
            {
               printf("   d= %g\n", energy);
               energy = fold(string,str2);
               printf("NO_MFE: %s\n", str2);
            }
            else
               printf("\n");
          }
      }

      if (pf)
      {
         if (!(mfe && give_up && (energy>0)))
         {
            /* unless we gave up in the mfe part */
            double prob, min_en, sfact=1.07;

            /* get a reasonable pf_scale */
            min_en = fold(string,str2);
            pf_scale = exp(-(sfact*min_en)/kT/struct_len);
            init_pf_fold(struct_len);

            energy = inverse_pf_fold(string);
            prob = exp(-energy/kT);
            hd = hamming(rstart, string);
            printf("PF:     %s  %3d  (%g)\n\n", string, hd, prob);
            free_pf_arrays();
         }
         if (! (mfe))
            found--;
      }
      free(string);
   }
   free(rstart);
   free_arrays();
   free(str2);

   printf("\n");
   //****************************************************
   //****************************************************
   return 0;
}



